# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ``Dimensions`` class."""

from __future__ import annotations

from typing import Union

from ansys.units import BaseDimensions


class Dimensions:
    """
    A representation of an arbitrary number of dimensions, where each dimension is a
    pair consisting of a base dimension and exponent.

    A dictionary of ``BaseDimensions`` and exponent is required
    for a non-dimensionless object.

    If any keys are duplicated in ``copy_from`` and ``dimensions`` then the
    associated values from ``dimensions`` are used.

    Parameters
    ----------
    dimensions : dict, optional
        Dictionary of {``BaseDimensions``: exponent, ...}.
    system : UnitSystem, optional
        The unit system for the dimensions.
    copy_from : Dimensions, optional
        A previous instance of Dimensions.
    """

    def __init__(
        self,
        dimensions: dict[BaseDimensions, Union[int, float]] = None,
        copy_from: Dimensions = None,
    ):
        dimensions = dimensions or {}
        self._dimensions = {
            **(copy_from._dimensions if copy_from else {}),
            **(dimensions),
        }

        for x, y in dimensions.items():
            if not isinstance(x, BaseDimensions):
                raise IncorrectDimensions()
            if y == 0:
                del self._dimensions[x]

    def _to_string(self):
        """
        Creates a string representation of the dimensions.

        Returns
        -------
        str
            A string version of the dimensions.
        """
        return str({x.name: y for x, y in self})

    def __str__(self):
        return self._to_string()

    def __repr__(self):
        return self._to_string()

    def __iter__(self):
        for item in self._dimensions.items():
            yield item

    def __mul__(self, other):
        results = self._dimensions.copy()
        for dim, value in other:
            if dim in results:
                results[dim] += value
            else:
                results[dim] = value
        return Dimensions(results)

    def __truediv__(self, other):
        results = self._dimensions.copy()
        for dim, value in other:
            if dim in results:
                results[dim] -= value
            else:
                results[dim] = -value
        return Dimensions(results)

    def __pow__(self, __value):
        results = self._dimensions.copy()
        for item in self:
            results[item[0]] *= __value
        return Dimensions(results)

    def __eq__(self, __value):
        dims = __value._dimensions.copy()
        for dim, value in self:
            if dim in dims:
                dims[dim] -= value
            else:
                return False
        if [False for v in dims.values() if v != 0]:
            return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __bool__(self):
        return bool(self._dimensions)


class IncorrectDimensions(ValueError):
    """Raised on initialization if a dimension is not of type ``BaseDimensions``."""

    def __init__(self):
        super().__init__("The `dimensions` key must be a 'BaseDimensions' object")



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Pyunits is a pythonic interface for units, unit systems, and unit conversions."""

try:
    import importlib.metadata as importlib_metadata
except ModuleNotFoundError:
    import importlib_metadata

__version__ = importlib_metadata.version(__name__.replace(".", "-"))

import os

from ansys.units.base_dimensions import BaseDimensions
from ansys.units.dimensions import Dimensions
from ansys.units.quantity import Quantity, get_si_value
from ansys.units.quantity_dimensions import QuantityDimensions
from ansys.units.systems import UnitSystem
from ansys.units.unit import Unit
from ansys.units.unit_registry import UnitRegistry
from ansys.units.variable_descriptor import (
    ConversionStrategy,
    MappingConversionStrategy,
    VariableCatalog,
    VariableDescriptor,
)

_THIS_DIRNAME = os.path.dirname(__file__)
_README_FILE = os.path.normpath(os.path.join(_THIS_DIRNAME, "docs", "README.rst"))

if os.path.exists(_README_FILE):
    with open(_README_FILE, encoding="utf8") as f:
        __doc__ = f.read()


__all__ = [
    "BaseDimensions",
    "Dimensions",
    "Quantity",
    "get_si_value",
    "UnitSystem",
    "Unit",
    "UnitRegistry",
    "QuantityDimensions",
    "VariableDescriptor",
    "VariableCatalog",
    "ConversionStrategy",
    "MappingConversionStrategy",
]



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ``UnitSystem`` class."""

from __future__ import annotations

from ansys.units import BaseDimensions
from ansys.units._constants import _base_units, _unit_systems


class UnitSystem:
    """
    A class representing base units for a unit system.

    Predefined unit systems work automatically and are configured when the
    package is initialized, whereas you can add user-defined systems at any time.

    Parameters
    ----------
    base_units: dict, optional
        Units mapped to base dimensions types.
    unit_sys: str, Unit, optional
        Predefined unit system.
    copy_from: UnitSystem, optional
        Make a copy of a unit system.

    Attributes
    ----------
    MASS
    LENGTH
    TIME
    TEMPERATURE
    TEMPERATURE_DIFFERENCE
    ANGLE
    CHEMICAL_AMOUNT
    LIGHT
    CURRENT
    SOLID_ANGLE
    """

    def __init__(
        self,
        base_units: dict[BaseDimensions, any] = None,
        system: str = None,
        copy_from: UnitSystem = None,
    ):
        if copy_from:
            self._units = copy_from._units
        else:
            if not system:
                system = "SI"
            if system not in _unit_systems:
                raise InvalidUnitSystem(system)
            else:
                self._units = _unit_systems[system].copy()

        if base_units:
            for unit_type, unit in base_units.items():
                self._units[unit_type.name] = unit

        for unit_type in BaseDimensions:
            unit = self._units[unit_type.name]
            self._set_type(unit_type=unit_type, unit=unit)

    def update(self, base_units: dict[BaseDimensions:any]):
        """
        Change the units of the unit system.

        Parameters
        ----------
        base_units: dict
            Units mapped to base dimensions types.
        """
        for unit_type, unit in base_units.items():
            self._set_type(unit_type=unit_type, unit=unit)

    def _set_type(self, unit_type: BaseDimensions, unit: any):
        """
        Checks that the unit is compatible with the unit type before being set.

        Parameters
        ----------
        unit_type: obj
            Unit system type slot for the new unit.
        unit: str, obj
            The unit to be assigned.
        """
        name = getattr(unit, "name", None) or unit

        if name not in _base_units:
            raise NotBaseUnit(name)
        if _base_units[name]["type"] != unit_type.name:
            raise IncorrectUnitType(name, unit_type)

        setattr(self, f"_{unit_type.name}", unit)

    @property
    def MASS(self):
        """Mass unit of the unit system."""
        return self._MASS

    @MASS.setter
    def MASS(self, new_unit):
        self._set_type(unit_type=BaseDimensions.MASS, unit=new_unit)

    @property
    def LENGTH(self):
        """Length unit of the unit system."""
        return self._LENGTH

    @LENGTH.setter
    def LENGTH(self, new_unit):
        self._set_type(unit_type=BaseDimensions.LENGTH, unit=new_unit)

    @property
    def TIME(self):
        """Time unit of the unit system."""
        return self._TIME

    @TIME.setter
    def TIME(self, new_unit):
        self._set_type(unit_type=BaseDimensions.TIME, unit=new_unit)

    @property
    def TEMPERATURE(self):
        """Temperature unit of the unit system."""
        return self._TEMPERATURE

    @TEMPERATURE.setter
    def TEMPERATURE(self, new_unit):
        self._set_type(unit_type=BaseDimensions.TEMPERATURE, unit=new_unit)

    @property
    def TEMPERATURE_DIFFERENCE(self):
        """Temperature unit of the unit system."""
        return self._TEMPERATURE_DIFFERENCE

    @TEMPERATURE_DIFFERENCE.setter
    def TEMPERATURE_DIFFERENCE(self, new_mass):
        self._set_type(unit_type=BaseDimensions.TEMPERATURE_DIFFERENCE, unit=new_mass)

    @property
    def ANGLE(self):
        """Angle unit of the unit system."""
        return self._ANGLE

    @ANGLE.setter
    def ANGLE(self, new_mass):
        self._set_type(unit_type=BaseDimensions.ANGLE, unit=new_mass)

    @property
    def CHEMICAL_AMOUNT(self):
        """Chemical Amount unit of the unit system."""
        return self._CHEMICAL_AMOUNT

    @CHEMICAL_AMOUNT.setter
    def CHEMICAL_AMOUNT(self, new_mass):
        self._set_type(unit_type=BaseDimensions.CHEMICAL_AMOUNT, unit=new_mass)

    @property
    def LIGHT(self):
        """Light unit of the unit system."""
        return self._LIGHT

    @LIGHT.setter
    def LIGHT(self, new_mass):
        self._set_type(unit_type=BaseDimensions.LIGHT, unit=new_mass)

    @property
    def CURRENT(self):
        """Current unit of the unit system."""
        return self._CURRENT

    @CURRENT.setter
    def CURRENT(self, new_mass):
        self._set_type(unit_type=BaseDimensions.CURRENT, unit=new_mass)

    @property
    def SOLID_ANGLE(self):
        """Solid Angle unit of the unit system."""
        return self._SOLID_ANGLE

    @SOLID_ANGLE.setter
    def SOLID_ANGLE(self, new_mass):
        self._set_type(unit_type=BaseDimensions.SOLID_ANGLE, unit=new_mass)

    def __repr__(self):
        units = ""
        for unit_type in BaseDimensions:
            unit = getattr(self, f"_{unit_type.name}")
            units += f"{unit_type.name}: {unit}\n"
        return units

    def __eq__(self, other_sys):
        for attr, value in self.__dict__.items():
            if getattr(other_sys, attr) != value:
                return False
        return True


class NotBaseUnit(ValueError):
    """Raised when a unit system unit is not a configured base unit."""

    def __init__(self, unit):
        super().__init__(
            f"`{unit}` is not a base unit. To use `{unit}`, add it to the "
            "`base_units` table within the cfg.yaml file."
        )


class InvalidUnitSystem(ValueError):
    """Raised when a unit system is initialized with an unsupported unit system."""

    def __init__(self, sys):
        super().__init__(f"`{sys}` is not a supported unit system.")


class IncorrectUnitType(ValueError):
    """Raised when a unit is provided that does not have a valid type of base unit."""

    def __init__(self, unit, unit_type):
        super().__init__(
            f"The unit `{unit}` is incompatible with unit system type: `{unit_type.name}`"
        )



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""Provides the ``Quantity`` class."""

from __future__ import annotations

from collections.abc import Iterable
import operator
from typing import Union

from ansys.units import BaseDimensions, Dimensions
from ansys.units.systems import UnitSystem
from ansys.units.unit import Unit

try:
    import numpy as np

    _array = np
except ImportError:
    _array = None

try:
    from matplotlib.units import AxisInfo, ConversionInterface, registry

    _ci = ConversionInterface
    _ai = AxisInfo
    _registry = registry
except ModuleNotFoundError:
    _ci, _ai, _registry = object, None, dict()


class Quantity:
    """
    A class representing a physical quantity's value and associated units.

    A Quantity object can be instantiated from a NumPy array or list only if the
    ``NumPy`` package is installed.

    The value argument is not required when using ``copy_from``. A value can be given
    to override the value from the copy.

    Implicit conversion from Quantity to float is only allowed if the operand is
    of quantity type angle or dimensionless.

    Parameters
    ----------
    value : int, float, list, np.array
        Real value of the quantity.
    units : str, Unit, optional
        Initializes the quantity's units using a string or ``Unit`` instance.
    quantity_table : dict[str, int], optional
        Initializes the quantity's units using the quantity table.
    dimensions : Dimensions, optional
        Initializes the quantity's units in SI using a ``Dimensions`` instance.
    copy_from : Quantity, optional
        An existing ``Quantity`` instance.

    Attributes
    ----------
    value
    units
    dimensions
    is_dimensionless
    """

    _chosen_units = []

    def __init__(
        self,
        value: Union[int, float] = None,
        units: Union[Unit, str] = None,
        quantity_table: dict = None,
        dimensions: Dimensions = None,
        copy_from: Quantity = None,
    ):
        if (
            (units and quantity_table)
            or (units and dimensions)
            or (quantity_table and dimensions)
        ):
            raise ExcessiveParameters()

        self._value = value
        self.unit = units

        if copy_from:
            if value:
                units = copy_from.units
            else:
                units = copy_from.units
                value = copy_from.value
        elif value is None:
            raise InsufficientArguments()

        if not isinstance(value, (float, int)):
            if isinstance(value, str):
                raise TypeError("value should be either float, int or [float, int].")
            if _array:
                if isinstance(value, _array.ndarray):
                    self._value = value
                else:
                    self._value = _array.array(value)
            elif not _array:
                raise NumPyRequired()
        else:
            self._value = float(value)

        if quantity_table:
            units = Unit(table=quantity_table)

        if dimensions:
            units = Unit(dimensions=dimensions)

        if not isinstance(units, Unit):
            units = Unit(units)

        if (
            (units.name in ["K", "R"] and value < 0)
            or (units.name == "C" and value < -273.15)
            or (units.name == "F" and value < -459.67)
        ):
            units = Unit(f"delta_{units.name}")

        self._unit = units

        for unit in self._chosen_units:
            if unit.name != units.name and self.dimensions == unit.dimensions:
                self._value = self.to(unit).value
                self._unit = unit

    @classmethod
    def preferred_units(
        cls, units: list[Union[Unit, str]], remove: bool = False
    ) -> None:
        """
        Add or remove preferred units.

        Quantities are automatically converted to preferred units when the
        quantity is initialized. Conversion is always carried out if the base
        units are consistent with the preferred units.

        Each preferred unit must have unique dimensions. To override units with
        the same dimensions, the original must first be removed.

        Parameters
        ----------
        units : list
            A list of units to be added or removed.
        remove : bool
            Specify if the units should be removed.
        """
        for unit in units:
            if isinstance(unit, str):
                unit = Unit(units=unit)
            if remove and unit in cls._chosen_units:
                cls._chosen_units.remove(unit)
            elif not remove:
                for chosen_unit in cls._chosen_units:
                    if chosen_unit.dimensions == unit.dimensions:
                        raise RequiresUniqueDimensions(unit, chosen_unit)
                cls._chosen_units.append(unit)

    @property
    def value(self):
        """Value in contained units."""
        return self._value

    @property
    def units(self) -> Unit:
        """The quantity's units."""
        return self._unit

    @property
    def dimensions(self):
        """The quantity's dimensions."""
        return self._unit.dimensions

    @property
    def is_dimensionless(self):
        """True if the quantity is dimensionless."""
        return not bool(self.dimensions)

    def to(self, to_units: Union[Unit, str]) -> "Quantity":
        """
        Perform quantity conversions.

        Parameters
        ----------
        to_units : Unit or str
            Desired unit to convert to.

        Returns
        -------
        Quantity
            Quantity instance in the desired units.

        Examples
        --------
        >>> speed_si = Quantity(value=5, units="m s^-1")
        >>> speed_bt = speed_si.to("ft s^-1")
        """

        if not isinstance(to_units, Unit):
            to_units = Unit(units=to_units)

        # Retrieve all SI required SI data and perform conversion
        new_value = (
            get_si_value(self) / to_units.si_scaling_factor
        ) - to_units.si_offset

        if self.dimensions != to_units.dimensions:
            raise IncompatibleDimensions(from_unit=self.units, to_unit=to_units)

        return Quantity(value=new_value, units=to_units)

    def compatible_units(self) -> set[str]:
        """
        Get all units with the same dimensions.

        Returns
        -------
        set
            A set of unit objects.
        """

        return self.units.compatible_units()

    def convert(self, system: UnitSystem) -> Quantity:
        """
        Convert a quantity into the unit system.

        Parameters
        ----------
        system : UnitSystem
            Unit system to convert to.

        Returns
        -------
        Quantity
            Quantity object converted into the unit system.

        Examples
        --------
        >>> ur = UnitRegistry()
        >>> speed_si = Quantity(value=5, units= ur.m / ur.s)
        >>> bt = UnitSystem(system="BT")
        >>> speed_bt = speed_si.convert(bt)
        """
        new_unit = self.units.convert(system)

        return self.to(to_units=new_unit)

    def _relative_unit_check(
        self, other, r_add_sub: bool, op: operator = operator.add
    ) -> Quantity:
        """
        Checks relative units for temperature differences.

        Parameters
        ----------
        other : float, int, Quantity
            The value to be added or subtracted to self.
        r_add_sub : bool
            Whether called in __radd__ or __rsub__.
        op : operator, optional
            The operation being performed. Default is addition.

        Returns
        -------
        Quantity
            Quantity instance changed to or from relative units.
        """
        if not isinstance(other, Quantity):
            other = Quantity(other)

        # Checks the temperatures at the unit level.
        new_units, other_units = op(self.units, other.units) or (
            self.units,
            other.units,
        )

        # If value does not equal relative units, use the corrected absolute units.
        if other.units.dimensions != other_units.dimensions:
            value = other.to(new_units).value
        # If both values are temperatures, use the corrected relative units.
        elif other.units.dimensions != self.units.dimensions:
            value = other.to(other_units).value
        else:
            value = (
                other.to(other_units).value if r_add_sub else other.to(self.units).value
            )

        new_value = op(self.value, value)
        if r_add_sub:
            return Quantity(value=new_value, units=other_units)
        return Quantity(value=new_value, units=new_units)

    def __float__(self):
        base_dims = BaseDimensions
        dims = Dimensions
        if self.dimensions in [
            dims(),
            dims(dimensions={base_dims.ANGLE: 1.0}),
            dims(dimensions={base_dims.SOLID_ANGLE: 1.0}),
        ]:
            return get_si_value(self)

    def __getitem__(self, idx):
        if isinstance(self.value, Iterable):
            return Quantity(value=float(self.value[idx]), units=self.units)

    def __str__(self):
        return f'({self.value}, "{self._unit.name}")'

    def __repr__(self):
        return f'Quantity ({self.value}, "{self._unit.name}")'

    def __pow__(self, other):
        new_value = self.value**other
        new_units = self._unit**other
        return Quantity(value=new_value, units=new_units)

    def __mul__(self, other):
        if isinstance(other, Quantity):
            new_value = self.value * other.value
            new_units = self._unit * other._unit
            return Quantity(
                value=new_value,
                units=new_units,
            )
        if isinstance(other, Unit):
            base_quantity = Quantity(1, other)
            return self * base_quantity

        if isinstance(other, (float, int)):
            return Quantity(value=self.value * other, units=self.units)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        if isinstance(other, Quantity):
            new_value = self.value / other.value
            new_units = self._unit / other._unit
            return Quantity(
                value=new_value,
                units=new_units,
            )

        if isinstance(other, Unit):
            base_quantity = Quantity(1, other)
            return self / base_quantity

        if isinstance(other, (float, int)):
            return Quantity(value=self.value / other, units=self._unit)

    def __rtruediv__(self, other):
        return Quantity(other, "") / self

    def __add__(self, other):
        return self._relative_unit_check(other, r_add_sub=False)

    def __radd__(self, other):
        return self._relative_unit_check(other, r_add_sub=True)

    def __sub__(self, other):
        return self._relative_unit_check(other, r_add_sub=False, op=operator.sub)

    def __rsub__(self, other):
        return self._relative_unit_check(other, r_add_sub=True, op=operator.sub)

    def __neg__(self):
        return Quantity(-self.value, self._unit)

    def validate_matching_dimensions(self, other):
        """Validates dimensions of quantities."""
        if isinstance(other, Quantity) and (self.dimensions != other.dimensions):
            raise IncompatibleDimensions(from_unit=self.units, to_unit=other.units)
        elif (
            (not self.is_dimensionless)
            and (not isinstance(other, Quantity))
            and isinstance(other, (float, int))
        ):
            raise IncompatibleQuantities(self, other)

    def _compute_single_value_comparison(self, other, op: operator):
        """Compares quantity values."""
        return (
            op(get_si_value(self), other)
            if self.is_dimensionless
            else op(get_si_value(self), get_si_value(other))
        )

    def __gt__(self, other):
        self.validate_matching_dimensions(other)
        return self._compute_single_value_comparison(other, op=operator.gt)

    def __ge__(self, other):
        self.validate_matching_dimensions(other)
        return self._compute_single_value_comparison(other, op=operator.ge)

    def __lt__(self, other):
        self.validate_matching_dimensions(other)
        return self._compute_single_value_comparison(other, op=operator.lt)

    def __le__(self, other):
        self.validate_matching_dimensions(other)
        return self._compute_single_value_comparison(other, op=operator.le)

    def __eq__(self, other):
        self.validate_matching_dimensions(other)
        if all(
            (
                isinstance(self.value, float),
                any(
                    (
                        isinstance(x, float)
                        for x in (other, getattr(other, "value", None))
                    )
                ),
            )
        ):
            return self._compute_single_value_comparison(other, op=operator.eq)
        # no type-checking here since array_equal happily processes anything
        return _array and _array.array_equal(get_si_value(self), get_si_value(other))

    def __ne__(self, other):
        return not self.__eq__(other)


def get_si_value(quantity: Quantity) -> float:
    """Returns a quantity's value in SI units."""

    def _convert(value, offset, factor):
        return float((value + offset) * factor)

    if isinstance(quantity.value, float):
        return _convert(
            quantity.value, quantity.units.si_offset, quantity.units.si_scaling_factor
        )
    if _array and isinstance(quantity.value, _array.ndarray):
        offset = quantity.units.si_offset
        factor = quantity.units.si_scaling_factor
        return _array.array([_convert(x, offset, factor) for x in quantity.value])


class ExcessiveParameters(ValueError):
    """Raised when excessive parameters are provided."""

    def __init__(self):
        super().__init__(
            "Quantity only accepts one of the following parameters: \
            (units) or (quantity_table) or (dimensions)."
        )


class InsufficientArguments(ValueError):
    """Raised when insufficient arguments are provided."""

    def __init__(self):
        super().__init__("Requires at least one 'value' or 'copy_from' argument.")


class IncompatibleDimensions(ValueError):
    """Raised when dimensions are incompatible."""

    def __init__(self, from_unit, to_unit):
        super().__init__(
            f"`{from_unit.name}` and `{to_unit.name}` have incompatible dimensions."
        )


class IncompatibleValue(ValueError):
    """Raised when an incompatible value is provided."""

    def __init__(self, value):
        super().__init__(f"`{value}` is incompatible with the current quantity object.")


class IncompatibleQuantities(ValueError):
    """Raised when quantities are incompatible."""

    def __init__(self, q1, q2):
        super().__init__(f"'{q1}' and '{q2}' are incompatible.")


class NumPyRequired(ModuleNotFoundError):
    """Raised when NumPy is unavailable."""

    def __init__(self):
        super().__init__("To use NumPy arrays and lists install NumPy.")


class InvalidFloatUsage(FloatingPointError):
    """Raised when float is unsupported for given type of quantity."""

    def __init__(self):
        super().__init__(
            "Only dimensionless quantities and angles can be used as a float."
        )


class RequiresUniqueDimensions(ValueError):
    """Raised when two units with the same dimensions are added to the chosen units."""

    def __init__(self, unit, other_unit):
        super().__init__(
            f"For '{unit.name}' to be added '{other_unit.name}' must be removed."
        )


class QuantityConverter(_ci):

    @staticmethod
    def convert(value, unit, axis):
        if isinstance(value, Quantity):
            return value._value
        else:
            return [quantity._value for quantity in value]

    @staticmethod
    def axisinfo(unit, axis):
        return _ai(label=unit)

    @staticmethod
    def default_units(x, axis):
        "Return the default unit for x or None"
        if isinstance(x, Quantity):
            attr = getattr(x, "unit", None)
        else:
            attr = getattr(x[0], "unit", None)
        return attr.name if isinstance(attr, Unit) else attr


_registry[Quantity] = QuantityConverter()



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.
# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Package initialization for ansys.units.variable_descriptor.

Exports the VariableCatalog catalog and core interfaces for working with
physical quantities in a product-agnostic way.

Example
-------

.. code-block:: python

    from ansys.units.variable_descriptor import VariableCatalog
    from ansys.fluent.core.variable_strategies import FluentSVarNamingStrategy

    quantity = VariableCatalog.PRESSURE
    strategy = FluentSVarNamingStrategy()

    print(strategy.to_string(quantity))
"""

from .strategy import ConversionStrategy, MappingConversionStrategy
from .variable_descriptor import VariableCatalog, VariableDescriptor

__all__ = [
    "VariableDescriptor",
    "VariableCatalog",
    "ConversionStrategy",
    "MappingConversionStrategy",
]



# Copyright (C) 2023 - 2025 ANSYS, Inc. and/or its affiliates.

# SPDX-License-Identifier: MIT
#
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.
"""
Defines immutable objects representing SI base and derived physical dimensions.

This module provides globally defined `Dimensions` instances corresponding to the
seven SI base quantities (e.g., mass, time) and many commonly used derived quantities
(e.g., force, energy, pressure). These dimension definitions are used internally
for dimensional analysis, quantity construction, and validation.

Each dimension is represented as a `Dimensions` object, which captures the
dimensional exponents of base quantities such as mass, length, and time.

This module is not intended for direct modification at runtime. All objects
are statically defined and immutable.

Examples
--------
>>> from ansys.units.dimensions import Dimensions
>>> from ansys.units.quantity_dimensions import QuantityDimensions

>>> force = QuantityDimensions.FORCE
>>> pressure = QuantityDimensions.PRESSURE
>>> pressure == force / QuantityDimensions.AREA
True
"""

from ansys.units.base_dimensions import BaseDimensions
from ansys.units.dimensions import Dimensions

# Local function to convert from a `BaseDimensions` enum object to a `Dimensions` object


def _make_base_dimensions(dimension: BaseDimensions) -> Dimensions:
    return Dimensions(dimensions={dimension: 1.0})


def _expand_vector_quantities(cls):
    angle = _make_base_dimensions(BaseDimensions.ANGLE)
    for name in getattr(cls, "_vector_quantities", []):
        value = getattr(cls, name)
        # cartesian only added by default
        setattr(cls, f"{name}_X", value)
        setattr(cls, f"{name}_Y", value)
        setattr(cls, f"{name}_Z", value)
        # magnitude
        setattr(cls, f"{name}_MAGNITUDE", value)
    return cls


@_expand_vector_quantities
class QuantityDimensions:
    """
    Defines immutable dimension objects for standard physical quantities.

    This class encapsulates SI base dimensions (e.g., MASS, TIME) and
    commonly used derived dimensions (e.g., FORCE, PRESSURE, ENERGY),
    using `Dimensions` instances.

    These objects support arithmetic operations (multiplication, division,
    exponentiation) to enable dimensional composition and analysis.

    All attributes are class-level constants and should be treated as immutable.

    Examples
    --------
    >>> QuantityDimensions.VELOCITY
    Dimensions(dimensions={BaseDimensions.LENGTH: 1.0, BaseDimensions.TIME: -1.0})
    """

    # Dimensions of base quantities
    MASS = _make_base_dimensions(BaseDimensions.MASS)
    LENGTH = _make_base_dimensions(BaseDimensions.LENGTH)
    TIME = _make_base_dimensions(BaseDimensions.TIME)
    TEMPERATURE = _make_base_dimensions(BaseDimensions.TEMPERATURE)
    CHEMICAL_AMOUNT = _make_base_dimensions(BaseDimensions.CHEMICAL_AMOUNT)
    LIGHT = _make_base_dimensions(BaseDimensions.LIGHT)
    CURRENT = _make_base_dimensions(BaseDimensions.CURRENT)

    # Supplementary base quantities
    ANGLE = _make_base_dimensions(BaseDimensions.ANGLE)
    SOLID_ANGLE = _make_base_dimensions(BaseDimensions.SOLID_ANGLE)
    TEMPERATURE_DIFFERENCE = _make_base_dimensions(
        BaseDimensions.TEMPERATURE_DIFFERENCE
    )

    # Dimensions of derived quantities

    # Kinematics and mechanics
    AREA = LENGTH**2
    VOLUME = LENGTH**3
    FREQUENCY = TIME**-1
    VELOCITY = LENGTH / TIME
    ACCELERATION = VELOCITY / TIME
    FORCE = MASS * ACCELERATION
    PRESSURE = FORCE / AREA
    STRESS = PRESSURE
    JERK = LENGTH / TIME**3
    SNAP = LENGTH / TIME**4
    ANGULAR_VELOCITY = TIME**-1
    ANGULAR_ACCELERATION = TIME**-2
    VOLUMETRIC_FLOW = VOLUME / TIME
    MOMENTUM = MASS * VELOCITY
    ANGULAR_MOMENTUM = LENGTH**2 * MASS / TIME
    TORQUE = LENGTH**2 * MASS / TIME**2
    ENERGY = LENGTH**2 * MASS / TIME**2
    POWER = ENERGY / TIME
    DENSITY = MASS / VOLUME
    SPECIFIC_VOLUME = VOLUME / MASS
    ENERGY_DENSITY = ENERGY / VOLUME
    SURFACE_TENSION = FORCE / LENGTH
    KINEMATIC_VISCOSITY = AREA / TIME
    DYNAMIC_VISCOSITY = PRESSURE * TIME
    MASS_FLOW_RATE = MASS / TIME
    MASS_CONTROL = MASS / TIME**3
    FREQUENCY_DRIFT = TIME**-2
    YANK = FORCE / TIME
    WAVENUMBER = LENGTH**-1
    AREA_DENSITY = MASS / AREA
    ACTION = ENERGY * TIME
    SPECIFIC_ENERGY = ENERGY / MASS
    HEAT_FLUX_DENSITY = POWER / AREA
    SURFACE_HEAT_FLUX = HEAT_FLUX_DENSITY
    LINEAR_MASS_DENSITY = MASS / LENGTH
    RADIANCE = POWER / (SOLID_ANGLE * AREA)
    SPECTRAL_POWER = POWER / LENGTH
    ABSORBED_DOSE_RATE = AREA / TIME**3
    FUEL_EFFICIENCY = LENGTH**-2
    SPECTRAL_IRRADIANCE = POWER / VOLUME
    ENERGY_FLUX_DENSITY = POWER / AREA
    COMPRESSIBILITY = PRESSURE**-1
    RADIANT_EXPOSURE = ENERGY / AREA
    MOMENT_OF_INERTIA = MASS * AREA
    SPECIFIC_ANGULAR_MOMENTUM = ANGULAR_MOMENTUM / MASS
    RADIANT_INTENSITY = POWER / SOLID_ANGLE
    SPECTRAL_INTENSITY = POWER / (SOLID_ANGLE * LENGTH)
    STRAIN_RATE = TIME**-1
    AXIAL_VELOCITY = VELOCITY
    RADIAL_VELOCITY = VELOCITY
    TANGENTIAL_VELOCITY = VELOCITY

    # Electromagnetics
    ELECTRIC_CHARGE = TIME * CURRENT
    ELECTRICAL_POTENTIAL = MASS * AREA * TIME**-3 / CURRENT
    ELECTRICAL_CAPACITANCE = MASS**-1 * LENGTH**-2 * TIME**4 * CURRENT**2
    ELECTRICAL_INDUCTANCE = MASS * LENGTH**2 * TIME**-2 * CURRENT**-2
    ELECTRICAL_RESISTANCE = ELECTRICAL_POTENTIAL / CURRENT
    ELECTRICAL_IMPEDANCE = ELECTRICAL_RESISTANCE
    ELECTRICAL_REACTANCE = ELECTRICAL_RESISTANCE
    ELECTRICAL_CONDUCTANCE = ELECTRICAL_RESISTANCE**-1
    MAGNETIC_FLUX = MASS * LENGTH**2 * TIME**-2 * CURRENT**-1
    MAGNETIC_INDUCTION = MASS * TIME**-2 * CURRENT**-1
    MAGNETIC_PERMEABILITY = LENGTH * MASS * TIME**-2 * CURRENT**-2
    MAGNETIC_VECTOR_POTENTIAL = LENGTH * MASS * TIME**-2 * CURRENT**-1

    # Chemistry
    MOLARITY = CHEMICAL_AMOUNT / VOLUME
    MOLAR_VOLUME = VOLUME / CHEMICAL_AMOUNT
    MOLAR_HEAT_CAPACITY = ENERGY / (TEMPERATURE * CHEMICAL_AMOUNT)
    MOLAR_ENTROPY = MOLAR_HEAT_CAPACITY
    MOLAR_ENERGY = ENERGY / CHEMICAL_AMOUNT
    MOLALITY = CHEMICAL_AMOUNT / MASS
    MOLAR_MASS = MASS / CHEMICAL_AMOUNT
    CATALYTIC_EFFICIENCY = VOLUME / (CHEMICAL_AMOUNT * TIME)

    # Photometry
    LUMINOUS_FLUX = LIGHT * SOLID_ANGLE
    ILLUMINANCE = LUMINOUS_FLUX / AREA
    LUMINANCE = LIGHT / AREA
    LUMINOUS_EXPOSURE = ILLUMINANCE * TIME

    # Thermodynamics
    HEAT_CAPACITY = ENERGY / TEMPERATURE
    SPECIFIC_HEAT_CAPACITY = HEAT_CAPACITY / MASS
    THERMAL_CONDUCTIVITY = POWER / (LENGTH * TEMPERATURE)
    THERMAL_RESISTANCE = TEMPERATURE / POWER
    TEMPERATURE_GRADIENT = TEMPERATURE / LENGTH
    ENTHALPY = ENERGY
    ENTROPY = ENERGY / TEMPERATURE
    SPECIFIC_ENTHALPY = ENTHALPY / MASS
    SPECIFIC_ENTROPY = ENTROPY / MASS
    INTERNAL_ENERGY = ENERGY
    TOTAL_ENERGY = ENERGY
    SPECIFIC_TOTAL_ENERGY = TOTAL_ENERGY / MASS
    SPECIFIC_INTERNAL_ENERGY = INTERNAL_ENERGY / MASS
    TOTAL_TEMPERATURE = TEMPERATURE

    # Turbulence
    PRANDTL_NUMBER = Dimensions()
    EFFECTIVE_PRANDTL_NUMBER = Dimensions()
    EFFECTIVE_THERMAL_CONDUCTIVITY = THERMAL_CONDUCTIVITY
    EFFECTIVE_VISCOSITY = DYNAMIC_VISCOSITY
    PRODUCTION_OF_TURBULENT_KINETIC_ENERGY = MASS * LENGTH**-1 * TIME**-3
    SPECIFIC_DISSIPATION_RATE = TIME**-1
    TURBULENT_DISSIPATION_RATE = LENGTH**2 * TIME**3
    TURBULENT_INTENSITY = Dimensions()
    TURBULENT_VISCOSITY = DYNAMIC_VISCOSITY
    TURBULENT_VISCOSITY_RATIO = Dimensions()
    TURBULENT_REYNOLDS_NUMBER = Dimensions()
    TURBULENT_KINETIC_ENERGY = LENGTH**2 * TIME**-2
    WALL_Y_PLUS = Dimensions()
    WALL_Y_STAR = Dimensions()

    # Fluid pressure
    STATIC_PRESSURE = PRESSURE  # equivalent
    ABSOLUTE_PRESSURE = PRESSURE
    DYNAMIC_PRESSURE = PRESSURE
    TOTAL_PRESSURE = PRESSURE
    PRESSURE_COEFFICIENT = Dimensions()

    # Other fluid and related quantities
    VORTICITY = TIME**-1
    CONVECTIVE_COURANT_NUMBER = Dimensions()
    CELL_REYNOLDS_NUMBER = Dimensions()
    Q_CRITERION = TIME**-2
    NORMALIZED_Q_CRITERION = Dimensions()
    MESH_VELOCITY = VELOCITY
    SPECIFIC_TOTAL_ENTHALPY = SPECIFIC_ENTHALPY
    SURFACE_NUSSELT_NUMBER = Dimensions()
    SURFACE_STANTON_NUMBER = Dimensions()
    SKIN_FRICTION_COEFFICIENT = Dimensions()
    SURFACE_HEAT_TRANSFER_COEFFICIENT = POWER * LENGTH**-2 * TEMPERATURE**-1
    WALL_ADJACENT_HEAT_TRANSFER_COEFFICIENT = POWER * LENGTH**-2 * TEMPERATURE**-1
    WALL_TEMPERATURE = TEMPERATURE
    WALL_ADJACENT_TEMPERATURE = TEMPERATURE
    WALL_TEMPERATURE_THIN = TEMPERATURE
    WALL_SHEAR_STRESS = STRESS

    # Coordinates
    POSITION = LENGTH

    # List of vector quantities. The only impact
    # of not adding a vector quantity here is that
    # the code will not autogenerate attributes
    # <quantity>_X etc for that quantity.
    _vector_quantities = [
        "VELOCITY",
        "ACCELERATION",
        "VORTICITY",
        "POSITION",
        "MESH_VELOCITY",
        "MOMENTUM",
        "ANGULAR_MOMENTUM",
        "TORQUE",
        "FORCE",
        "WALL_SHEAR_STRESS",
    ]
